#include <cstdio>
#include <cstdlib>
#include <cmath>
#if defined(WIN32)
//#  pragma comment(linker, "/subsystem:\"windows\" /entry:\"mainCRTStartup\"")
#  pragma comment(lib, "glew32.lib")
#  include "glew.h"
#  include "glut.h"
#elif defined(__APPLE__) || defined(MACOSX)
#  include <GLUT/glut.h>
#else
#  define GL_GLEXT_PROTOTYPES
#  include <GL/glut.h>
#endif

/*
** 形状データ
*/
#include "Obj.h"
static const char objectData[] = "bunny_s.obj";
static Obj *obj;

/*
** シェーダ
*/
#include "shadersource.h"
static const char vertSource[] = "simple.vert";
static const char fragSource[] = "simple.frag";
static GLuint shader;

/*
** カメラ
*/
static const float cameraPosition[] = { 0.0f, 0.0f, 200.0f };
static const float cameraTarget[] = { 0.0f, 0.0f, 0.0f };
static const float cameraUp[] = { 0.0f, 1.0f, 0.0f };
static const float cameraFov = 60.0f;
static const float cameraNear = 1.0f;
static const float cameraFar = 1000.0f;

/*
** ウィンドウサイズ
*/
static int width, height;

/*
** 視線の向き
*/
static void lookat(const float *e, const float *t, const float *u, float *m)
{
  float tx, ty, tz, l;

  /* z 軸 = e - t */
  tx = e[0] - t[0];
  ty = e[1] - t[1];
  tz = e[2] - t[2];
  l = sqrt(tx * tx + ty * ty + tz * tz);
  if (l > 0.0f) {
    m[6] = tx / l;
    m[7] = ty / l;
    m[8] = tz / l;
  }

  /* x 軸 = u x z 軸 */
  tx = u[1] * m[8] - u[2] * m[7];
  ty = u[2] * m[6] - u[0] * m[8];
  tz = u[0] * m[7] - u[1] * m[6];
	l = sqrt(tx * tx + ty * ty + tz * tz);
  if (l > 0.0f) {
    m[0] = tx / l;
    m[1] = ty / l;
    m[2] = tz / l;
  }

  /* y 軸 = z 軸 x x 軸 */
  m[3] = m[7] * m[2] - m[8] * m[1];
  m[4] = m[8] * m[0] - m[6] * m[2];
  m[5] = m[6] * m[1] - m[7] * m[0];
}

/*
** 画面表示
*/
static void display(void)
{
  /* 開始時刻の取得 */
  int beginTime = glutGet(GLUT_ELAPSED_TIME);
  
  /* 描画開始 */
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glEnable(GL_DEPTH_TEST);

  /* シェーダプログラムを有効にする */
  glUseProgram(shader);

  /* 視点と投影面との距離を 1 としたときの投影面の大きさ */
  GLfloat sy = tan(cameraFov * 0.5f * 3.1415926f / 180.0f);
  GLfloat sx = sy * (float)width / (float)height;

  /* 投影面の前方面上の投影像のスケール */
  GLfloat org[3];
  org[2] = cameraNear;
  org[0] = org[2] * sx;
  org[1] = org[2] * sy;

  /* 投影面の後方面上の投影像の前方面上の投影像からの変位 */
  GLfloat dir[3];
  dir[2] = cameraFar - cameraNear;
  dir[0] = dir[2] * sx;
  dir[1] = dir[2] * sy;

  /* 視線の方向を回転する行列 */
  GLfloat rot[9];
  lookat(cameraPosition, cameraTarget, cameraUp, rot);

  /* バーテックスシェーダの uniform 変数の場所を得る */
  GLint camLocation = glGetUniformLocation(shader, "cam");
  GLint orgLocation = glGetUniformLocation(shader, "org");
  GLint dirLocation = glGetUniformLocation(shader, "dir");
  GLint rotLocation = glGetUniformLocation(shader, "rot");

  /* バーテックスシェーダの uniform 変数に値を設定する */
  glUniform3fv(camLocation, 1, cameraPosition);
  glUniform3fv(orgLocation, 1, org);
  glUniform3fv(dirLocation, 1, dir);
  glUniformMatrix3fv(rotLocation, 1, GL_FALSE, rot);

  /* 表示領域いっぱいに描くポリゴンの頂点座標 */
  static const GLfloat scr[][3] = {
    { -1.0f, -1.0f, -1.0f },
    {  1.0f, -1.0f, -1.0f },
    {  1.0f,  1.0f, -1.0f },
    { -1.0f,  1.0f, -1.0f },
  };

  /* 頂点バッファオブジェクトを生成する */
  GLuint bufferObject;
  glGenBuffers(1, &bufferObject);

  /* 頂点バッファオブジェクトにデータを送る */
  glBindBuffer(GL_ARRAY_BUFFER, bufferObject);
  glBufferData(GL_ARRAY_BUFFER, sizeof scr, scr, GL_STATIC_DRAW);

  /* attribute 変数の場所を得る */
  GLint scrLocation = glGetAttribLocation(shader, "scr");

  /* attribute 変数にバッファオブジェクトの場所を指定する */
  glVertexAttribPointer(scrLocation, sizeof scr[0] / sizeof scr[0][0], GL_FLOAT, GL_FALSE, 0, 0);

  /* attribute 変数を有効にする */
  glEnableVertexAttribArray(scrLocation);

  /* フラグメントシェーダの uniform 変数の場所を得る */
  GLint posLocation = glGetUniformLocation(shader, "pos");
  GLint vecLocation = glGetUniformLocation(shader, "vec");

  /* ポリゴンの描画 */
  obj->draw(posLocation, vecLocation, sizeof scr / sizeof scr[0]);

  /* attribute 変数を無効にする */
  glDisableVertexAttribArray(scrLocation);

  /* 頂点バッファオブジェクトを削除する */
  glDeleteBuffers(1, &bufferObject);

  /* シェーダプログラムを無効にする */
  glUseProgram(0);

  /* 描画終了 */
  glDisable(GL_DEPTH_TEST);
  glFinish();

  /* 経過時間の表示 */
  fprintf(stderr, "Elapsed time = %f\n", (double)(glutGet(GLUT_ELAPSED_TIME) - beginTime) * 0.001);
}

/*
** ウィンドウサイズ
*/
static void resize(int w, int h)
{
  /* ウィンドウの幅と高さを保存する */
  width = w;
  height = h;
  
  /* ウィンドウ全体をビューポートにする */
  glViewport(0, 0, w, h);
}

/*
** キーボード
*/
static void keyboard(unsigned char key, int x, int y)
{
  switch (key) {
    /* q か Q か ESC をタイプしたら終了 */
    case '\033':
    case 'q':
    case 'Q':
      exit(0);
    case 'g':
    case 'G':
      /* g か G をタイプしたら再描画 */
      glutPostRedisplay();
    default:
      break;
  }
}

/*
** シェーダプログラムの読み込み
*/
static GLuint loadShader(const char *vertSource, const char *fragSource)
{
  /* シェーダオブジェクトの作成 */
  GLuint vertShader = glCreateShader(GL_VERTEX_SHADER);
  GLuint fragShader = glCreateShader(GL_FRAGMENT_SHADER);
  
  /* シェーダのソースプログラムの読み込み */
  if (readShaderSource(vertShader, vertSource)) return 0;
  if (readShaderSource(fragShader, fragSource)) return 0;
  
  /* シェーダプログラムのコンパイル／リンク結果を得る変数 */
  GLint status;
  
  /* バーテックスシェーダのソースプログラムのコンパイル */
  glCompileShader(vertShader);
  glGetShaderiv(vertShader, GL_COMPILE_STATUS, &status);
  printShaderInfoLog(vertShader);
  if (status == GL_FALSE) {
    fprintf(stderr, "Compile error in %s.\n", vertSource);
    return 0;
  }
  
  /* フラグメントシェーダのソースプログラムのコンパイル */
  glCompileShader(fragShader);
  glGetShaderiv(fragShader, GL_COMPILE_STATUS, &status);
  printShaderInfoLog(fragShader);
  if (status == GL_FALSE) {
    fprintf(stderr, "Compile error in %s.\n", fragSource);
    return 0;
  }
  
  /* プログラムオブジェクトの作成 */
  GLuint gl2Program = glCreateProgram();
  
  /* シェーダオブジェクトのシェーダプログラムへの登録 */
  glAttachShader(gl2Program, vertShader);
  glAttachShader(gl2Program, fragShader);
  
  /* シェーダオブジェクトの削除 */
  glDeleteShader(vertShader);
  glDeleteShader(fragShader);
  
  /* シェーダプログラムのリンク */
  glLinkProgram(gl2Program);
  glGetProgramiv(gl2Program, GL_LINK_STATUS, &status);
  printProgramInfoLog(gl2Program);
  if (status == GL_FALSE) {
    fprintf(stderr, "Link error.\n");
    return 0;
  }
  
  return gl2Program;
}

/*
** 後始末
*/
static void cleanup(void)
{
  delete obj;
}

/*
** 初期化
*/
static void init(void)
{
#if defined(WIN32)
  /* GLEW の初期化 */
  GLenum err = glewInit();
  if (err != GLEW_OK) {
    fprintf(stderr, "Error: %s\n", glewGetErrorString(err));
    exit(1);
  }
#endif
  /* シェーダプログラムの読み込み */
  shader = loadShader(vertSource, fragSource);
  if (shader == 0) exit(1);
  
  /* 形状データの読み込み */
  obj = new Obj;
  if (obj == 0) exit(1);
  atexit(cleanup);
  if (!obj->load(objectData)) exit(1);

  /* 背景色 */
  glClearColor(0.0, 0.1, 0.3, 1.0);
  
  /* 経過時間の初期化 */
  glutGet(GLUT_ELAPSED_TIME);
}

/*
** メインプログラム
*/
int main(int argc, char *argv[])
{
  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH);
  glutCreateWindow(argv[0]);
  glutDisplayFunc(display);
  glutReshapeFunc(resize);
  glutKeyboardFunc(keyboard);
  init();
  glutMainLoop();
  
  return 0;
}
